
  include 'mkl_vsl.fi'
  
  module szcluster

  use tinkermass
  
  implicit none
  
  real ode_omm,ode_omk,ode_oml,ode_w
  
  integer, parameter :: MAXNFIELD = 50
  integer, parameter :: MAXNCATALOG = 512

  type szcatalog
     character(LEN=Ini_max_string_len) cat_name
     character(LEN=Ini_max_string_len), dimension(1:MAXNFIELD) :: falserate_files
     integer nfield
     integer, dimension(1:MAXNFIELD) :: nper
     integer, dimension(1:MAXNFIELD) :: firsti
     real*8, dimension(1:MAXNFIELD) :: szscale
     real*8, dimension(1:MAXNFIELD) :: area
     real*8, dimension(1:MAXNFIELD) :: fracarea
     integer n
     real*8, dimension(0:MAXNCATALOG-1) :: sz_vec,x_vec,z_vec,&
          z_err_vec, da_vec, m_vec,logx_err_vec
     integer ntx,nmg
     real*8, dimension(:,:), allocatable:: r_mg,r_tx,mgr_vec,&
          txr_vec
     
  end type szcatalog
  
  type(szcatalog) :: cat
  
  integer, parameter :: mxarrsize = 1024
  type szparam
     real area
     real delta
     integer rho_crit, grid2d
     integer nm,nz,nx,nsz
     real*8, dimension(0:mxarrsize-1) ::  m_arr, z_arr, x_arr, sz_arr
     real*8, dimension(0:mxarrsize-1,1:MAXNFIELD) :: falserates
     real*8 :: lm_min,m_min, delta_lm, delta_z, z_min,z_max,delta_sz, sz_min, sz_max
     real*8 :: x_min,x_max,delta_x, z_cut,m_cut,mpivot_sz,z_pivot, szthresh
     real*8, dimension(0:3) :: x_central_vals, x_prior
     real*8, dimension(0:5) :: sz_central_vals, sz_prior
     real*8, allocatable,dimension(:,:) :: sig_rz
  end type szparam
  type(szparam) :: szparams
  
  type cosmoprm
     integer nk,nz
     real*8, dimension(0: num_matter_power-1) :: k_arr
     real*8, dimension(0: matter_power_lnzsteps-1) :: z_arr
     real*8, dimension(0: num_matter_power-1,0:matter_power_lnzsteps-1) :: mp_arr
     integer :: dozinterp
     real*8 :: z_min,z_max,dlnz
     real*8 :: k_min,k_max
     real :: h0,w,ombh2, omch2, omb, omc
     real :: omm, oml, omk, a_002, n_s, tau
  end type cosmoprm
  type(cosmoprm) :: csm
  
  type scalingreln
     real :: ax, bx, cx, dx
     real :: asz, bsz, csz, dsz, esz, fsz
     real :: rho

  end type scalingreln
  type(scalingreln) :: sr

  type massfunc
     integer nm
     real*8 m_min,m_max
     integer nz
     real*8 z_min,z_max
     real*8, dimension(:,:), pointer :: dn 
  end type massfunc
  type(massfunc) mf

  
  real*8, dimension(:,:), allocatable :: grid3dtmp
  !last index will be field
  real*8, dimension(:,:,:), allocatable :: gridexpected
  real*8, dimension(:,:,:,:), allocatable :: grid3dexpected
  real*8, parameter :: max_z_error = 10d0
contains  
  !functions go here
  !testing routine
  subroutine SetCSM(omm,oml,omk,w)
    real*8 :: omm,oml,omk,w
    csm%omm=omm
    csm%oml=oml
    csm%omk=omk
    csm%w=w
  end subroutine SetCSM

  subroutine InitCalCosmoParamsCatalog(param_file,cat_file)
    character(*) :: param_file,cat_file
    integer n,i,j,k,ii,jj,kk,j0,j1
    real*8, dimension(:,:),allocatable :: tmp_r_mg,tmp_mgr_vec,tmp_r_tx,tmp_txr_vec
    integer, dimension(0:MAXNCATALOG-1) :: saveindices
    real*8, dimension(0:3000) :: tmpfalse,xfalse
    character(LEN=Ini_max_string_len) tmpstr
    integer, save :: initialized =0
    if (initialized .ne. 0) then 
       write(*,*)'Double call to InitCalCosmoParamsCatalog'
       call DoStop
    endif
    initialized =1

    call OpenTxtFile(param_file, tmp_file_unit)
    read (tmp_file_unit,*) szparams%area
    read (tmp_file_unit,*) szparams%delta
    read (tmp_file_unit,*) szparams%rho_crit
    read (tmp_file_unit,*) szparams%grid2d
    read (tmp_file_unit,*) szparams%nm
    if (szparams%nm .gt. mxarrsize) then
       write(*,*) 'Error with nm size'
       call DoStop
    end if
    do i=0,szparams%nm-1
       read (tmp_file_unit,*) szparams%m_arr(i)
    end do
    read (tmp_file_unit,*) szparams%lm_min
    read (tmp_file_unit,*) szparams%m_min
    read (tmp_file_unit,*) szparams%delta_lm
    read (tmp_file_unit,*) szparams%nz
    if (szparams%nz .gt. mxarrsize) then
       write(*,*) 'Error with nz size'
       call DoStop
    end if
    do i=0,szparams%nz-1
       read (tmp_file_unit,*) szparams%z_arr(i)
    end do
    read (tmp_file_unit,*) szparams%delta_z
    read (tmp_file_unit,*) szparams%z_min
    read (tmp_file_unit,*) szparams%z_max

    read (tmp_file_unit,*) szparams%nsz
    if (szparams%nsz .gt. mxarrsize) then
       write(*,*) 'Error with nsz size'
       call DoStop
    end if
    do i=0,szparams%nsz-1
       read (tmp_file_unit,*) szparams%sz_arr(i)
    end do
    read (tmp_file_unit,*) szparams%delta_sz
    read (tmp_file_unit,*) szparams%sz_min
    read (tmp_file_unit,*) szparams%sz_max
    
    read (tmp_file_unit,*) szparams%nx
    if (szparams%nx .gt. mxarrsize) then
       write(*,*) 'Error with nm size'
       call DoStop
    end if
    do i=0,szparams%nx-1
       read (tmp_file_unit,*) szparams%x_arr(i)
    end do
    read (tmp_file_unit,*) szparams%delta_x
    read (tmp_file_unit,*) szparams%x_min
    read (tmp_file_unit,*) szparams%x_max
    do i=0,3
       read (tmp_file_unit,*) szparams%x_central_vals(i)
    end do
    do i=0,5
       read (tmp_file_unit,*) szparams%sz_central_vals(i)
    end do
    do i=0,3
       read (tmp_file_unit,*) szparams%x_prior(i)
    end do
    do i=0,5
       read (tmp_file_unit,*) szparams%sz_prior(i)
    end do
    read (tmp_file_unit,*) szparams%z_cut
    read (tmp_file_unit,*) szparams%m_cut
    read (tmp_file_unit,*) szparams%mpivot_sz
    read (tmp_file_unit,*) szparams%z_pivot
    read (tmp_file_unit,*) szparams%szthresh
    close(tmp_file_unit)

    allocate(szparams%sig_rz(0:szparams%nm-1,0:szparams%nz-1))
    call OpenTxtFile(cat_file, tmp_file_unit)
    read (tmp_file_unit,*) cat%cat_name
    read (tmp_file_unit,*) n, k
    if (n .gt. MAXNCATALOG) then
       write(*,*)'truncating catalog to: ',MAXNCATALOG
       n = MAXNCATALOG
    end if
    if (k .gt. MAXNFIELD) then
       write(*,*)'stopping! truncating catalog Nfield to: ',MAXNFIELD
       call DoStop
    end if
    cat%nfield = k
    read (tmp_file_unit,*)cat%nper(1:k)
    read (tmp_file_unit,*)cat%area(1:k)
    cat%fracarea(1:k)=cat%area(1:k)/szparams%area
    read (tmp_file_unit,*)cat%szscale(1:k)
    if (n .ne. sum(cat%nper(1:k))) then
       write(*,*)'catalog number mismatch, quitting!',n,cat%nper(1:k)
       call DoStop
    endif
    
    do i=1,k 
       read (tmp_file_unit,'(A128)') tmpstr
       cat%falserate_files(i)=trim(tmpstr)
    enddo
    
    j=0
    kk=0
    do i=1,k 
      cat%firsti(i)=j
      jj=0
      do ii=1,cat%nper(i) 
         read (tmp_file_unit,*) cat%sz_vec(j),cat%x_vec(j),&
              cat%logx_err_vec(j),&
              cat%z_vec(j),cat%z_err_vec(j),&
              cat%da_vec(j),cat%m_vec(j)
         
         if (cat%z_vec(j) .gt. szparams%z_cut)   then
            saveindices(j)=kk
            j=j+1
            jj=jj+1
         endif
         kk=kk+1
      enddo
      if (jj .lt. 1) then
         write(*,*)'stopping! not set up to handle <1 cluster per field (postcuts)',jj
         call DoStop
      endif
      cat%nper(i) =jj
   enddo
   cat%n = j
   if (cat%n .ne. sum(cat%nper(1:k))) then
      write(*,*)'catalog number mismatch2, quitting!',cat%n,cat%nper(1:k)
      call DoStop
   endif

    read (tmp_file_unit,*) cat%nmg,cat%ntx
    if (cat%nmg .gt. 0 .and. cat%ntx .gt. 0) then
       allocate(cat%r_mg(0:cat%nmg-1,0:cat%n-1), cat%mgr_vec(0:cat%nmg-1,0:cat%n-1))
       allocate(cat%r_tx(0:cat%ntx-1,0:cat%n-1), cat%txr_vec(0:cat%ntx-1,0:cat%n-1))
       allocate(tmp_r_mg(0:cat%nmg-1,0:n-1), tmp_mgr_vec(0:cat%nmg-1,0:n-1))
       allocate(tmp_r_tx(0:cat%ntx-1,0:n-1), tmp_txr_vec(0:cat%ntx-1,0:n-1))
       do j=0,cat%nmg-1
          do i=0,n-1
             read (tmp_file_unit,*) tmp_r_mg(j,i), tmp_mgr_vec(j,i)
          enddo
       enddo
       do j=0,cat%ntx-1
          do i=0,n-1
             read (tmp_file_unit,*) tmp_r_tx(j,i), tmp_txr_vec(j,i)
          enddo
       enddo
       cat%r_mg(0:cat%nmg-1,0:cat%n-1) = tmp_r_mg(0:cat%nmg-1,saveindices(0:cat%n-1))
       cat%mgr_vec(0:cat%nmg-1,0:cat%n-1) = tmp_mgr_vec(0:cat%nmg-1,saveindices(0:cat%n-1))
       cat%r_tx(0:cat%ntx-1,0:cat%n-1) = tmp_r_tx(0:cat%ntx-1,saveindices(0:cat%n-1))
       cat%txr_vec(0:cat%ntx-1,0:cat%n-1) = tmp_txr_vec(0:cat%ntx-1,saveindices(0:cat%n-1))
       deallocate(tmp_r_mg,tmp_mgr_vec,tmp_r_tx,tmp_txr_vec)
    endif
    close(tmp_file_unit)

    do i=1,k 
       call OpenTxtFile(cat%falserate_files(i), tmp_file_unit)
       do ii=0,3000
          xfalse(ii)=ii*.02
          read (tmp_file_unit,*) tmpfalse(ii)
       enddo
       close(tmp_file_unit)
       jj=0
       j0=floor((szparams%sz_arr(0))/0.02)+1
       do ii=0,szparams%nsz-2
         j1=floor((szparams%sz_arr(ii+1))/0.02)+1
          szparams%falserates(ii,i) = sum(tmpfalse(j0:j1-1))
          j0=j1
       enddo
       szparams%falserates(szparams%nsz-1,i)=0d0

       !make per field instead of per deg2
       szparams%falserates(0:szparams%nsz-1,i) = szparams%falserates(0:szparams%nsz-1,i) * cat%area(i)

       where(szparams%falserates(0:szparams%nsz-1,i) < 0d0)
          szparams%falserates(0:szparams%nsz-1,i) = 0d0
       endwhere
    enddo

    csm%nk = num_matter_power
    do i=0,num_matter_power-1
       csm%k_arr(i) = exp(log(matter_power_minkh)+(i)*matter_power_dlnkh)
    enddo
    csm%k_min = csm%k_arr(0)
    csm%k_max = csm%k_arr(num_matter_power-1)

    csm%nz = matter_power_lnzsteps
    csm%dlnz = log(matter_power_maxz+1) / (matter_power_lnzsteps -1 + 1e-13)
    do i=0,matter_power_lnzsteps-1
       csm%z_arr(i) = exp(csm%dlnz*(i))-1.0
    enddo
    csm%z_arr(0)=0.0
    csm%z_min = csm%z_arr(0)
    csm%z_max = csm%z_arr(matter_power_lnzsteps-1)

    !do checks for useable z info
    csm%dozinterp=1
    if ( ( matter_power_lnzsteps .le. 1) .or. &
         (szparams%z_min .lt. csm%z_min) .or. &
         (szparams%z_max .gt. csm%z_max) ) csm%dozinterp=0

    !and prep mass function structure
    mf%nm = szparams%nm
    mf%nz = szparams%nz
    allocate( mf%dn(0:mf%nm-1,0:mf%nz-1))
    mf%m_min=szparams%m_arr(0)
    mf%m_max=szparams%m_arr(mf%nm-1)
    mf%z_min=szparams%z_arr(0)
    mf%z_max=szparams%z_arr(mf%nz-1)


    if (szparams%grid2d .eq. 1) then
       allocate(gridexpected(0:szparams%nsz-1,0:szparams%nz-1,cat%nfield))
    else
       allocate(grid3dexpected(0:szparams%nx-1,0:szparams%nsz-1,0:szparams%nz-1,cat%nfield))
    end if

  end subroutine INITCALCOSMOPARAMSCATALOG
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !comput cosmological growth function using ODE
  subroutine calcosmo_growth_a_derivs(x,y,dydx)
    real*8, dimension(2) :: dydx
    real*8, dimension(2), intent(in) :: y
    real*8, intent(in) :: x
    real*8 :: w_prime, a, w_a, qevol, h2evol, dlnh_dlna
    real*8 :: omega_m_a
    w_prime=0.0	!evolution of w set to 0 (constant)
    a=exp(x)
    w_a=ode_w+(1.d0-a)*w_prime	!w(a) == w0
    qevol=a**(-3.0*(1.0 +ode_w+w_prime))*exp(-3.0*w_prime*(1.0 - a))
    h2evol=ode_omm*a**(-3.0) + ode_oml*qevol + ode_omk*a**(-2.0)
    dlnh_dlna=-0.5/h2evol*&
         (3.0*ode_omm*a**(-3.0) + 3.0*ode_oml*qevol*(1.+w_a) - &
         2.0*ode_omk*a**(-2.0) )
    omega_m_a=ode_omm/h2evol * a**(-3.0)
    
    dydx(1)=y(2)
    dydx(2)=-(4.0 +dlnh_dlna)*y(2)-(3.0 +dlnh_dlna-1.5*omega_m_a)*y(1)
  end subroutine calcosmo_growth_a_derivs
  

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function calcosmo_growth_ode(z)
    real*8, intent(in) :: z
    real*8, parameter :: omk=0.0, eps=0.05
    real*8 :: a, a_start, x_start
    real*8,dimension(2) :: y_start != (/ 1.0, 0.0 /)

    real*8 :: h1,d,hwant
    real*8 :: calcosmo_growth_ode
    INTEGER nbad,nok,KMAXX,NMAX,count,fail
    PARAMETER (NMAX=50,KMAXX=200)
    real*8, dimension(KMAXX) :: xp,yptmp
    real*8, dimension(NMAX,KMAXX) :: yp
    real*8, dimension(1) :: xi,yi
    hwant = 0.2d0
    ode_omm=csm%omm
    ode_oml=csm%oml
    ode_omk=csm%omk
    ode_w=csm%w
    
    y_start(1)=1.0
    y_start(2)=0.0
    a=1./(1.+z)
    a_start=a
    if (a .gt. 1.1e-3) a_start=1.1e-3  
    x_start=log(a_start)
    h1=hwant
    fail=0
    call dodeint(y_start,2,x_start,0d0,eps,h1,0d0,hwant,nok,nbad,calcosmo_growth_a_derivs,xp,yp,count,fail)
    if (fail .eq. 1) then
       calcosmo_growth_ode=-1d30
       return
    end if
 
    yptmp(1:count) = yp(1,1:count)
     xi(1) = log(a)
    call dlinterpol(yptmp,xp,count, xi,yi,1)
    calcosmo_growth_ode=yi(1)*a/y_start(1)
  end function calcosmo_growth_ode
  

!------------------------------------------------------------------------------
! NAME:
!       CALCOSMO_GENERATE_MF.PRO
!
! PURPOSE:
!       TAKE A csm (COSMOLOGY) STRUCTURE AND RETURN A DN/DLOGMDZ (MASS FUNCTION)
!------------------------------------
  subroutine CALCOSMO_GENERATE_MF( FNL)
    real*8, intent(in) :: FNL
    real*8 :: rho_back, z, new_delta, sigl,sigh
    integer i, sig_kcut,j,iz0
    real*8 :: lnz,ilnz,f
    real*8, dimension(0:szparams%nm-1) :: dsdm_arr, sig_arr, rm_arr,&
         growthsig_arr, growthdsdm_arr, fsig, r_arr, g1,sig_0
    real*8, dimension(0:szparams%nm-1, 0:szparams%nz-1) ::  dn1
    real*8 :: cvol,zmax
    real*8, parameter ::  cH0 = 2997.92458
    real*8 :: domega,dlm, rm, growth_fact, var1, mfm0
    real*8 :: dndmdz, dz, int_dz, da_calc0,s8, dndlogmdz
    integer :: int_z_steps,fail,max_int_z_steps
    real*8, allocatable, dimension(:) :: int_z_arr, da_int
    real*8, dimension(0:csm%nk-1) :: kr, window,this_mp
    real*8 :: delta_k
    real*8, dimension(1) :: t1,t2

    if (csm%dozinterp .eq. 0 ) then 
       call     CALCOSMO_GENERATE_MF_NOPKZINTERP( FNL)
       return
    endif
    
    !zero it out
    szparams%sig_rz = 0.0

    domega = szparams%area * (PI/ 180d0)**2
    rho_back = 2.77536627e11*csm%omm
    mf%dn(:,:)=0
    dz=szparams%z_arr(1)-szparams%z_arr(0)
    dlm=log10(szparams%m_arr(1))-log10(szparams%m_arr(0))
    delta_k = log(csm%k_arr(1)) - log(csm%k_arr(0))

    !calculate comoving volume element
    int_dz=dz/10.
    zmax = szparams%z_arr(szparams%nz-1)
    max_int_z_steps=NINT(zmax/int_dz)
    allocate(int_z_arr(max_int_z_steps+1),da_int(max_int_z_steps+1))
    do i=0,max_int_z_steps
       int_z_arr(i+1)=i*int_dz
    end do
    da_int(:)=1.d0/sqrt(csm%omm*(1d0+int_z_arr(:))**3 + csm%oml*(1d0+int_z_arr(:))**(3d0+3d0*csm%w))

    rm_arr = (szparams%m_arr * (0.75d0/Pi/rho_back))**(1.d0/3.d0)
    do i=0,szparams%nz-1
       !we're going to use linear interpolation in logz space
       lnz = log(szparams%z_arr(i)+1)
       ilnz = lnz / csm%dlnz
       iz0 = floor(ilnz+ 1d-13)
       f = ilnz-iz0
       this_mp = (1-f)*csm%mp_arr(:,iz0) + f*csm%mp_arr(:,iz0+1) 
       do j=0,szparams%nm-1
          kr(:)=csm%k_arr(:)* rm_arr(j)
          window(:) = 3.*(sin(kr(:))/kr(:)**3 - cos(kr(:))/kr(:)**2)
          szparams%sig_rz(j,i) = sqrt( (delta_k/(2d0*Pi**2))*&
               sum(&
               csm%k_arr(:)**3* &
               this_mp(:)*&
               window(:)*window(:)&
               ))
       enddo
    enddo

    do i=0,szparams%nz-1
       z = szparams%z_arr(i)

       new_delta=szparams%delta
       if (szparams%rho_crit .eq. 1) &
            new_delta=szparams%delta/((csm%omm*(1d0+z)**3)/(csm%omm*(1d0+z)**3+csm%oml*(1d0+z)**(3d0+3d0*csm%w)))
       
       growthsig_arr(:)  =szparams%sig_rz(:,i)

       call deriv(szparams%m_arr,growthsig_arr,growthdsdm_arr,szparams%nm)
       
       call tinker_f_sigma(growthsig_arr,szparams%nm,z,new_delta,fsig)
       
       !!calculate comoving volume element
       int_z_steps=NINT(z/int_dz)
       da_calc0=dint_tabulated_reg(int_z_arr(1:int_z_steps+1),da_int(1:int_z_steps+1),int_z_steps+1)
       
       cvol = (da_calc0/(1d0+z))**2/(sqrt(csm%omm*(1d0+z)**3 + csm%oml*(1d0+z)**(3d0+3d0*csm%w)))
       cvol = cvol * (1d0+z)**2 * cH0**3
       
       do j=0,szparams%nm-1 
          if (fsig(j) .ne. 0) then
             if (log(fsig(j)) + log(rho_back)-log(szparams%m_arr(j))-log(growthsig_arr(j))+log(-1d0*growthdsdm_arr(j)) .ge. -708d0) then 
                
                dndmdz = -1d0*fsig(j)*rho_back/szparams%m_arr(j)/&
                     growthsig_arr(j)*growthdsdm_arr(j)
                dndlogmdz = dndmdz * szparams%m_arr(j) * log(10d0)
                mf%dn(j,i) = dndlogmdz * cvol *domega *dlm *dz
                
             endif
          endif
       enddo

    end do !z-loop
    
    
    !apply semi-analytic non-gaussian approximation

    if (FNL .ne. 0) then 
       dn1(:,:)=0
       lnz = 0
       this_mp = csm%mp_arr(:,0) 
       do j=0,szparams%nm-1
          kr(:)=csm%k_arr(:)* rm_arr(j)
          window(:) = 3.*(sin(kr(:))/kr(:)**3 - cos(kr(:))/kr(:)**2)
          sig_0(j) = sqrt( (delta_k/(2d0*Pi**2))*&
               sum(&
               csm%k_arr(:)**3* &
               this_mp(:)*&
               window(:)*window(:)&
               ))
       enddo
       t1(1)=8d0
       call       dlinterpol(sig_0,rm_arr,szparams%nm,t1,t2,1)
       s8=t2(1)
       do i=0,szparams%nz-1 
          z=szparams%z_arr(i)
      
          do j=0,szparams%nm-1 

             
             mfm0=1.3d-4*FNL*s8*szparams%sig_rz(j,i)**(-2)+1d0
             var1 = 1.4d-4*(abs(FNL)*s8)**0.8 / szparams%sig_rz(j,i)
             r_arr(:)=szparams%m_arr(j)/szparams%m_arr(:)
             g1=exp(-0.5*(r_arr(:)-mfm0)**2/var1)
             dn1(j,i)=sum(g1(:)*mf%dn(:,i))/sum(g1(:))
          end do
          
       enddo
       mf%dn(:,:)=dn1(:,:)
    endif

    deallocate(int_z_arr,da_int)
  end subroutine CALCOSMO_GENERATE_MF
  
  subroutine CALCOSMO_GENERATE_MF_NOPKZINTERP( FNL)
    real*8, intent(in) :: FNL
    real*8, dimension(0:csm%nk-1) :: mp_splint2
    integer, parameter :: sig_n = 116
    real*8, dimension(0:sig_n-1) :: sig_splint0, sig_splint1, sig_splint2

    real*8 :: rho_back, z, new_delta, sigl,sigh
    integer i, sig_kcut,j
    real*8, dimension(0:szparams%nm-1) :: dsdm_arr, sig_arr, &
         growthsig_arr, growthdsdm_arr, fsig, r_arr, g1
    real*8, dimension(0:szparams%nm-1, 0:szparams%nz-1) ::  dn1
    real*8 :: cvol,zmax
    real*8, parameter ::  cH0 = 2997.92458
    real*8 :: domega,dlm, rm, growth_fact, var1, mfm0
    real*8 :: dndmdz, dz, int_dz, da_calc0,s8, dndlogmdz
    integer :: int_z_steps,fail,max_int_z_steps
    real*8, allocatable, dimension(:) :: int_z_arr, da_int
    real*8, dimension(0:csm%nk-1) :: kr, window, sint
    real*8 :: delta_k
    domega = szparams%area * (PI/ 180d0)**2
    
    mf%dn(:,:)=0
    dz=szparams%z_arr(1)-szparams%z_arr(0)
    dlm=log10(szparams%m_arr(1))-log10(szparams%m_arr(0))
    delta_k = log(csm%k_arr(1)) - log(csm%k_arr(0))

    !calculate comoving volume element
    int_dz=dz/10.
    zmax = szparams%z_arr(szparams%nz-1)
    max_int_z_steps=NINT(zmax/int_dz)
    allocate(int_z_arr(max_int_z_steps+1),da_int(max_int_z_steps+1))
    do i=0,max_int_z_steps
       int_z_arr(i+1)=i*int_dz
    end do
    da_int(:)=1.d0/sqrt(csm%omm*(1d0+int_z_arr(:))**3 + csm%oml*(1d0+int_z_arr(:))**(3d0+3d0*csm%w))

    !set up splint in order to calculate power spectrum from CAMB data
    call  dspline(csm%k_arr,csm%mp_arr(:,0),csm%nk,1.d30,1.d30,mp_splint2)

    rho_back = 2.77536627e11*csm%omm
    DO i=0,sig_n-1
       sig_splint0(i) = 10d0 ** (11.5d0 + 5d-2 * i)
       sig_splint0(i) =  (3.d0*sig_splint0(i)/(4.d0*Pi*rho_back))**(1.d0/3.d0)
       !this is also sig_rm

       kr(:)=csm%k_arr(:)* sig_splint0(i)
       window(:) = 3.*(sin(kr(:))/kr(:)**3 - cos(kr(:))/kr(:)**2)
       sint(0:csm%nk-1)=(1d0/(2d0*Pi**2))*csm%k_arr(:)**2*csm%mp_arr(:,0)*window(:)*window(:)

       sig_splint1(i)=sqrt(sum(csm%k_arr(0:csm%nk-1)*sint(0:csm%nk-1))*delta_k)

    end DO
    call  dspline(sig_splint0,sig_splint1,sig_n,1.d30,1.d30,sig_splint2)

    do i=0,szparams%nm-1
       rm = (3d0*0.99 * szparams%m_arr(i) / (4.d0*Pi*rho_back))**(1.d0/3.d0)
       call dsplint(sig_splint0,sig_splint1,sig_splint2,sig_n,rm,sigl,fail)
       rm = (3d0 * szparams%m_arr(i) / (4.d0*Pi*rho_back))**(1.d0/3.d0)
       call dsplint(sig_splint0,sig_splint1,sig_splint2,sig_n,rm,sig_arr(i),fail)
       rm = (3d0*1.01 * szparams%m_arr(i) / (4.d0*Pi*rho_back))**(1.d0/3.d0)
       call dsplint(sig_splint0,sig_splint1,sig_splint2,sig_n,rm,sigh,fail)
       dsdm_arr(i) = (sigh-sigl) / (.02*szparams%m_arr(i))
    end do


    !Calculate dndm via the Tinker mass function
    !Start z loop
    do i=0,szparams%nz-1
       z = szparams%z_arr(i)

       !if using rho_crit then redefine delta -> delta/omega_m(z)
       new_delta=szparams%delta
       if (szparams%rho_crit .eq. 1) &
            new_delta=szparams%delta/((csm%omm*(1d0+z)**3)/(csm%omm*(1d0+z)**3+csm%oml*(1d0+z)**(3d0+3d0*csm%w)))

       !Calculate growth factor only once per z iteration
       growth_fact=calcosmo_growth_ode(z)
       if (growth_fact .eq. -1d30) then 
          write(*,*)'failed growth ode, zeroing mass function'
          mf%dn(:,:)=0d0
          deallocate(int_z_arr,da_int)
          return
       endif
       growthsig_arr(:)  =growth_fact * sig_arr(:)
       growthdsdm_arr(:) = growth_fact * dsdm_arr(:)
       
       call tinker_f_sigma(growthsig_arr,szparams%nm,z,new_delta,fsig)
       
       !!calculate comoving volume element
       int_z_steps=NINT(z/int_dz)
       da_calc0=dint_tabulated_reg(int_z_arr(1:int_z_steps+1),da_int(1:int_z_steps+1),int_z_steps+1)

       cvol = (da_calc0/(1d0+z))**2/(sqrt(csm%omm*(1d0+z)**3 + csm%oml*(1d0+z)**(3d0+3d0*csm%w)))
       cvol = cvol * (1d0+z)**2 * cH0**3
       
       do j=0,szparams%nm-1 
          if (fsig(j) .ne. 0) then
             if (log(fsig(j)) + log(rho_back)-log(szparams%m_arr(j))-log(growthsig_arr(j))+log(-1d0*growthdsdm_arr(j)) .ge. -708d0) then 
                
                dndmdz = -1d0*fsig(j)*rho_back/szparams%m_arr(j)/&
                     growthsig_arr(j)*growthdsdm_arr(j)
                !Convert to dn/dlogM:
                dndlogmdz = dndmdz * szparams%m_arr(j) * log(10d0)
                mf%dn(j,i) = dndlogmdz * cvol *domega *dlm *dz
                
             endif
          endif
       enddo

    end do !z-loop
    
    
    !apply semi-analytic non-gaussian approximation
    if (FNL .ne. 0.0) then 
       call dsplint(sig_splint0,sig_splint1,sig_splint2,sig_n,8d0,s8,fail)
       dn1(:,:)=0
       do i=0,szparams%nz-1 
          z=szparams%z_arr(i)
          growth_fact=calcosmo_growth_ode(z)
          if (growth_fact .eq. -1d30) then 
             write(*,*)'failed growth ode, zeroing mass function'
             mf%dn(:,:)=0d0
             deallocate(int_z_arr,da_int)
             return
          endif
          growthsig_arr(:)  =growth_fact * sig_arr(:)
          
          do j=0,szparams%nm-1 
             mfm0=1.3d-4*FNL*s8*growthsig_arr(j)**(-2)+1d0
             var1 = 1.4d-4*(abs(FNL)*s8)**0.8 / growthsig_arr(j)
             r_arr(:)=szparams%m_arr(j)/szparams%m_arr(:)
             g1=exp(-0.5*(r_arr(:)-mfm0)**2/var1)
             dn1(j,i)=sum(g1(:)*mf%dn(:,i))/sum(g1(:))
          end do
       end do
       mf%dn(:,:)=dn1(:,:)
    endif	 
    deallocate(int_z_arr,da_int)
  end subroutine CALCOSMO_GENERATE_MF_NOPKZINTERP

    
  function Cluster_LnLike(CMB, Info,Nclus,clust_prior)
    use cmbtypes, only : pivot_k
    Type (CMBparams) CMB
    Type(ParamsetInfo) Info
    integer i,j
    real*8 fNL
    real, intent(out)::Nclus,clust_prior
    integer fail
    integer,parameter :: ioff=num_norm - num_sz_params + 1
    real*8 :: Cluster_LnLike
    Nclus = -1.23456789
    do j=0, matter_power_lnzsteps-1
       do i=0,csm%nk-1 
          csm%mp_arr(i,j) = Info%Theory%matter_power(i+1,j+1)
       enddo
    enddo
    csm%ombh2 = CMB%ombh2
    csm%omch2 = CMB%omch2
    csm%oml = CMB%omv
    csm%tau = CMB%reserved(1)
    csm%n_s = CMB%InitPower(1)
    csm%a_002 = CMB%norm(1)/1e10 * (0.002/pivot_k)**(csm%n_s-1d0)
    csm%w = CMB%w
    csm%h0 = CMB%H0
    csm%omb =  csm%ombh2 /(csm%h0/100.)**2
    csm%omc =  csm%omch2 /(csm%h0/100.)**2
    csm%omm = csm%omc + csm%omb
    write(*,*)'cmb ',CMB%ombh2,CMB%omch2,CMB%omv,CMB%H0
    write(*,*)'interp',csm%ombh2, csm%omch2,csm%oml
    write(*,*)'interp2',csm%h0,csm%omb, csm%omc,csm%omm


    print*,CMB%norm(ioff:ioff+10),0.0
    sr%asz = CMB%norm(ioff)
    sr%bsz = CMB%norm(ioff+1)
    sr%csz = CMB%norm(ioff+2)
    sr%dsz = CMB%norm(ioff+3)
    sr%esz = CMB%norm(ioff+4)
    sr%fsz = CMB%norm(ioff+5)

    sr%ax = CMB%norm(ioff+6)
    sr%bx = CMB%norm(ioff+7)
    sr%cx = CMB%norm(ioff+8)
    sr%dx = CMB%norm(ioff+9)
    sr%rho = CMB%norm(ioff+10)

    !disabled since only used in clusters, not other likelihoods
    fNL = 0.0 !CMB%norm(ioff+11)
    
    !generate mass function
    call calcosmo_generate_mf(fNL)

    do i=0,szparams%nm-1
       if (szparams%m_arr(i) .lt. szparams%m_cut) then
          mf%dn(i,:) = 0.0
       else
          exit
       end if
    end do
    do i=0,szparams%nz-1
       if (szparams%z_arr(i) .lt. szparams%z_cut) then
          mf%dn(:,i) = 0.0
       else
          exit
       end if
    end do
   
    !calculate log likelihood
    if (szparams%grid2d .eq. 1) then
       Cluster_LnLike = calcosmo_calculate_twod_lnlike(Nclus,clust_prior,fail)
    else 
       Cluster_LnLike = calcosmo_calculate_lnlike(Nclus,clust_prior,fail)
    end if
    if (fail .eq. 1)  then 
       Cluster_LnLike=  logZero
       clust_prior = logZero
       Nclus = 0.0
    endif
    
  end function Cluster_LnLike
  

!------------------------------------------------------------------------------
! NAME:
!       CALCOSMO_GENERATE_2D_GRID.PRO
!
! PURPOSE:
!       GENERATE A 2D EXPECTATION GRID, TO BE USED IN POISSON LIKELIHOOD
!------------------------------------

SUBROUTINE CALCOSMO_GENERATE_TWOD_GRID()
  use MKL_VSL
  use MKL_VSL_TYPE
  integer, save :: firsttime = 1
  integer, parameter :: MAXSIZE = 1024

  real*8, dimension(0:MAXSIZE-1), save :: lnszdb,  alogm
  real*8, dimension(0:szparams%nm-1) :: cen_lnmsz, censz,ocensz
  real*8, dimension(0:szparams%nsz-1) :: junknsz, junk2nsz
  real*8, save :: sz_binsize, x_binsize, z_binsize
  real*8, parameter :: sigma = 5d0  ! IMPORTANT APPROXIMATION PARAMETER
  real*8 :: expdsz , sigsz,  ll_sz, ul_sz, denom_exp_arg, renorm_fact
  integer, parameter :: nz_int = 1000
  real*8, dimension(0:nz_int) :: z_int_arr, iez_int_arr
  real*8, dimension(0:mf%nz-1) :: da_arr, ez_arr
  real*8 :: da_pivot, ez_pivot, sk, sz_sigma_obs, D_H
  integer :: i,j,k
  !conv task vars
  TYPE(VSL_CONV_TASK), save ::  task, task2
  integer start(1)
  real*8 :: x(9),h(9),y(18)
  integer ::  mode,xshape,yshape,zshape,xstride
  integer, save :: nkern,halfnkern
  integer :: status
  integer, parameter :: MAXNKERN = 1024
  real*8, save, dimension(MAXNKERN) :: ker, zout,zcor
  real*8, dimension(1) :: t1,t2

  !define 'de-boosted' sz vector
  if (firsttime .eq. 1) then
     if ((MAXSIZE .lt. szparams%nsz) .or. (MAXSIZE .lt. szparams%nm)) then
        write(*,*) 'maxsize too small',MAXSIZE, szparams%nsz,szparams%nm
        call DoStop
     endif
     firsttime = 0
     do i=0,szparams%nsz-1
        if (szparams%sz_arr(i) .gt. sqrt(7d0)) then
           lnszdb(i) = log(sqrt(szparams%sz_arr(i)**2 - 3d0))
        else
           lnszdb(i) = log(szparams%sz_arr(i))
        end if
     enddo
     sz_binsize = abs(szparams%sz_arr(1) - szparams%sz_arr(0))
     z_binsize = szparams%z_arr(1) - szparams%z_arr(0)
     alogm(0:szparams%nm-1) = log(szparams%m_arr(:)/szparams%mpivot_sz)

     sz_sigma_obs=1d0/sz_binsize
     halfnkern = floor(5*sz_sigma_obs)+1
     nkern = 2*halfnkern+1

     if (nkern + szparams%nz -1 .gt. MAXNKERN) then
        write(*,*)'MAXNKERN too small',nkern + szparams%nz -1, MAXNKERN
        call DoStop
     endif
     do i=-1*halfnkern, halfnkern
        ker(1+i+halfnkern) = exp(-0.5/sz_sigma_obs**2 * (i**2) )
     end do
     sk=sum(ker(1:nkern))
     ker(1:nkern) = ker(1:nkern)/sk
     zcor(halfnkern) =  ker(1)
     do i=2,halfnkern
        zcor(halfnkern+1 - i) = zcor(halfnkern+2 - i) + ker(i)
     end do

     mode = VSL_CONV_MODE_DIRECT;  
     status = vsldconvnewtaskx1d(task, mode, nkern, szparams%nsz, szparams%nsz, ker, 1)

     start(1) = halfnkern
     status = vslconvsetstart(task,start) 
     
  endif

  D_H=2997.92458/(csm%H0/100.0)		!Hubble distance, note factor of h!
  gridexpected(:,:,:) = 0


! precompute a bunch of things
  sigsz = 2d0*sr%fsz**2
  renorm_fact = sz_binsize / (sqrt(2d0*Pi)*sr%fsz)

  !create angular diameter distance array (note units here are MPC, not h**(-1)MPC)
  ez_arr(:) = sqrt( csm%omm*(1d0+szparams%z_arr(:))**3 + (csm%oml) * (1d0+szparams%z_arr(:))**(3d0*(1d0+(csm%w))))
  
  do i=0,szparams%nz-1 
     
     do j=0,nz_int
        z_int_arr(j) = j*(szparams%z_arr(i)/nz_int)
     end do
     
     iez_int_arr(:) = 1d0 / sqrt( csm%omm*(1d0 + z_int_arr(:))**3d0 + (csm%oml) * (1d0 + z_int_arr(:))**(3d0 * (1d0 + (csm%w)) ) )
     da_arr(i)=D_H*Dint_tabulated_reg(z_int_arr,iez_int_arr,nz_int+1)/(1.0+szparams%z_arr(i))

  end do


  

!calculate pivot points for E(z) and D_A(z)
  t1(1) = szparams%z_pivot
  call dsinterpol(ez_arr,szparams%z_arr,mf%nz, t1,t2, 1)
  ez_pivot = t2(1)
  call dsinterpol(da_arr,szparams%z_arr,mf%nz, t1,t2, 1)
  da_pivot = t2(1)

  junknsz(:) =  renorm_fact/szparams%sz_arr(:)
 
  
 !populate sz-z grid
  do i=0,mf%nz-1 
     if (maxval(mf%dn(:,i)) .eq. 0) cycle ! skip empty redshift bins
     cen_lnmsz(:) = log(sr%asz) + sr%bsz * alogm(0:szparams%nm-1) + sr%csz * log((1d0+szparams%z_arr(i))/(1d0+szparams%z_pivot))+ sr%dsz * log(ez_arr(i)/ez_pivot)+ sr%esz * log(da_arr(i)/da_pivot)
     
     do j=0,szparams%nsz-1 
        if (szparams%sz_arr(j) .lt. sqrt(7d0) .and. szparams%sz_arr(j) .ge. 2d0) then
           cycle
        endif

        do k=1,cat%nfield
           !TRY VECTORIZING TOO
           ocensz = lnszdb(j) - cen_lnmsz - log(cat%szscale(k))
           call vdsqr(szparams%nm,ocensz, censz)
           censz = (-1d0/sigsz) * censz 
           call vdexp(szparams%nm,censz,ocensz)
           call vdmul(szparams%nm,mf%dn(:,i),ocensz,censz)
           gridexpected(j,i,k) =  sum(censz) * cat%fracarea(k)
        enddo
     enddo

     do k=1,cat%nfield
        call vdmul(szparams%nsz,gridexpected(:,i,k),junknsz,junk2nsz)
        
        status = vsldconvexecx1d(task, junk2nsz, 1, zout, 1)
        zout(1:halfnkern) = zout(1:halfnkern) + &
             junk2nsz(0) * zcor(1:halfnkern)
        zout(szparams%nsz:szparams%nsz-halfnkern+1:-1) = &
             zout(szparams%nsz:szparams%nsz-halfnkern+1:-1) +&
             junk2nsz(szparams%nsz-1)* zcor(1:halfnkern)
        
        call vdpacki(szparams%nsz,zout,1,gridexpected(:,i,k))
     enddo
  enddo
END SUBROUTINE CALCOSMO_GENERATE_TWOD_GRID


!------------------------------------------------------------------------------
! NAME:
!       CALCOSMO_CALCULATE_2D_LNLIKE.PRO
!
! PURPOSE:
!       TAKE A CATALOG, SET OF COSMOLOGICAL/SCALING RELATION PARAMETERS AND
!       COMPUTE THE LOG-LIKELIHOOD ACCORDING TO POISSON STATISTICS
!------------------------------------
!note lnlike not multiplied by -1 until very end
function  CALCOSMO_CALCULATE_TWOD_LNLIKE(Nclus, clust_prior,FAIL)
  real*8 :: CALCOSMO_CALCULATE_TWOD_LNLIKE
  integer, intent(out) :: fail
  real, intent(out) :: Nclus,clust_prior
  real*8 :: data_like,poisson_lnlike,total_expected, prior_lnlike
  real*8, dimension(1) :: this_data_like
  real*8, parameter :: interp_fact = 10d0
  integer, parameter :: iinterp_fact = 10
  integer :: nzspline 
  real*8, dimension(0:(iinterp_fact * szparams%nz -1)) :: old_z_spline_arr_index,old_z_spline_arr,old_z_spline_temp
  integer :: i_clust, i,zl,zh,isz_index,dnz,zl0,zh0
  real*8 :: whcutoff
  real*8 :: delta_z, z_min, this_sz, sz_min,delta_sz,sz_index,ff,this_z,z_index
  
  real*8, dimension(0:(iinterp_fact*szparams%nz-1)) :: exp_z0, exp_z1, exp_z
  real*8, dimension(0:szparams%nz-1) :: zinds,zscr
  real*8, dimension(0:szparams%nsz-1) :: szinds
  real*8, dimension(1) :: t1,t2
  real*8, dimension(0:3) :: xpriorscr
  real*8, dimension(0:5) :: szpriorscr
  real*8 :: min_z_error
  
  integer ifield
  logical :: lowerlimit

  CALCOSMO_CALCULATE_TWOD_LNLIKE=-1d30

  nzspline =iinterp_fact * szparams%nz
  do i=0,szparams%nz-1
     zinds(i)=i
  end do
  do i=0,szparams%nsz-1
     szinds(i)=i
  end do

  fail = 0
! check for crazy scaling relations
  if ( sr%ax .lt. 0.001 .or. sr%ax .gt. 100. &
     .or. sr%bx .lt. 0.05 .or. sr%bx .gt. 100. &
     .or. sr%cx .lt. -10. .or. sr%cx .gt. 10. &
     .or. sr%dx .lt. 0.02 .or. sr%dx .gt. 4. &
     .or. sr%asz .lt. 0.001 .or. sr%asz .gt. 100. &
     .or. sr%bsz .lt. 0.05 .or. sr%bsz .gt. 100. &
     .or. sr%csz .lt. -10. .or. sr%csz .gt. 10. &
     .or. sr%fsz .lt. 0.02 .or. sr%fsz .gt. 4. &
     .or. sr%rho .lt. 0. .or. sr%rho .gt. 1. ) then 
     fail = 1
     return
  endif
  
  call calcosmo_generate_twod_grid()

  data_like = 0

  delta_sz = (szparams%sz_arr(szparams%nsz-1) - szparams%sz_arr(0)) / (szparams%nsz-1)
  sz_min = szparams%sz_arr(0)

  delta_z = (szparams%z_arr(szparams%nz-1) - szparams%z_arr(0)) / (szparams%nz-1) / interp_fact
  min_z_error = abs(delta_z  / sqrt(2d0 *Pi))
  z_min = szparams%z_arr(0)
  do i=0,nzspline-1
     old_z_spline_arr(i) = z_min + delta_z*(i)
  end do

  whcutoff = ((szparams%szthresh - sz_min) / delta_sz) 
  total_expected=0d0
  do ifield=1,cat%nfield
     do i_clust=cat%firsti(ifield),cat%firsti(ifield)+cat%nper(ifield)-1 
        lowerlimit = .false.
        if (cat%z_err_vec(i_clust) .le. min_z_error) then	!no z_error
           zl = min(max(floor((cat%z_vec(i_clust) - z_min)/delta_z),0),nzspline-1)
           zh = min(zl+1,nzspline-1)
        else
           if (cat%z_err_vec(i_clust) .ge. max_z_error) then
              zl = min(max(floor((cat%z_vec(i_clust) - z_min)/delta_z +0.5),0),nzspline-1)
              zh = nzspline-1
              lowerlimit = .true.
           else
              zl = min(max(floor((cat%z_vec(i_clust) - z_min - 5*cat%z_err_vec(i_clust))/delta_z),0),nzspline-1)
              zh = min(max(floor((cat%z_vec(i_clust) - z_min + 5*cat%z_err_vec(i_clust))/delta_z)+1,0),nzspline-1)
           endif
        end if
     
        if (zl .eq. zh) then 
           fail = 1
           write(*,*) 'Error, data outside of GRID!',zl,zh
           return
        endif
        
        ! 1) Evaluate the grid in dimension 2 (\xi space)
        this_sz = cat%sz_vec(i_clust)
        
        sz_index = (this_sz - sz_min) / delta_sz
        isz_index = floor(sz_index)
        
        zscr=gridexpected(isz_index,:,ifield)
        call dlinterpol(zscr,szparams%z_arr,szparams%nz,&
             old_z_spline_arr(zl:zh), exp_z0,zh-zl+1)
        zscr=gridexpected(isz_index+1,:,ifield)
        call dlinterpol(zscr,szparams%z_arr,szparams%nz,&
             old_z_spline_arr(zl:zh), exp_z1,zh-zl+1)
        
        ff =  (sz_index-isz_index)
        exp_z(0:zh-zl) = (1d0-ff)*exp_z0(0:zh-zl) + ff*exp_z1(0:zh-zl)
        exp_z(0:zh-zl) = exp_z(0:zh-zl) / interp_fact
     
        ! 2) Evaluate the grid in dimension 1 (z space)
        this_z = cat%z_vec(i_clust)
        z_index = (this_z - z_min) / delta_z - zl
        if (cat%z_err_vec(i_clust) .le. min_z_error) then
           ! no z error (spec-z, simulated data)
           t1(1) = z_index
           !Note that using zinds is ok here since zh-zl always =1 
           !we're not really using the full extent of exp_z
           call dlinterpol(exp_z(0:zh-zl),zinds(0:zh-zl),zh-zl+1,t1,t2,1)
           this_data_like = t2(1)
        else    
           if (lowerlimit) then 
              !summing everything up
              this_data_like(1) = sum(exp_z(0:zh-zl)) + szparams%falserates(isz_index,ifield)
           else
              dnz = floor(5*(cat%z_err_vec(i_clust)/delta_z))+1
              zl0 = max(floor(z_index-dnz),0)+zl
              zh0 = min(floor(z_index+dnz),zh-zl)+zl
              
              exp_z0(0:zh0-zl0) = (old_z_spline_arr(zl0:zh0) - this_z)/cat%z_err_vec(i_clust)
              exp_z0(0:zh0-zl0) = exp(-0.5 * (exp_z0(0:zh0-zl0))**2)
              exp_z0(0:zh0-zl0) = exp_z0(0:zh0-zl0) /sum(exp_z0(0:zh0-zl0))
              
              this_data_like(1) = sum(exp_z0(0:zh0-zl0) *exp_z(zl0-zl:zh0-zl))
        endif
        endif
        ! we can rescale by binsize, which only adds a constant to the lnlike
        !   however, the lnlike now becomes independent of binsize

        this_data_like = this_data_like/(delta_z*delta_sz)
        data_like = data_like + log(this_data_like(1))
        
     end do

     ! now use a fancy 3d integrator (canned routine)
     total_expected = total_expected + &
          square_int_twod(gridexpected(:,:,ifield),szinds,&
          szparams%nsz, zinds,&
          szparams%nz,&
          (/ 1d0*whcutoff, 1d0*szparams%nsz-1, 0d0, 1d0*szparams%nz-1 /), 96)
  enddo
  total_expected = total_expected+sum(szparams%falserates(floor(whcutoff+1d-6):szparams%nsz-1,1:cat%nfield))

  Nclus = total_expected
  poisson_lnlike = data_like - total_expected
  
! add X and SZ priors when applicable
  prior_lnlike=0.d0
  szpriorscr(:)=0
  xpriorscr(:)=0
  where (szparams%x_prior > 0.0)
     xpriorscr = -0.5 * ((szparams%x_central_vals - &
          (/ sr%ax,sr%bx,sr%cx,sr%dx /))/szparams%x_prior)**2
  endwhere
  where (szparams%sz_prior > 0.0)
     szpriorscr = -0.5 * ( (szparams%sz_central_vals - &
          (/ sr%asz,sr%bsz,sr%csz,sr%dsz,sr%esz,sr%fsz /) ) / &
          szparams%sz_prior)**2
  endwhere
  prior_lnlike = prior_lnlike + sum(xpriorscr) + sum(szpriorscr)
  clust_prior=-1*prior_lnlike
#ifdef _NOPRIOR_
  CALCOSMO_CALCULATE_TWOD_LNLIKE = -1*(poisson_lnlike)
#else
  CALCOSMO_CALCULATE_TWOD_LNLIKE = -1*(prior_lnlike + poisson_lnlike)
#endif 

END function CALCOSMO_CALCULATE_TWOD_LNLIKE

function  CALCOSMO_CALCULATE_LNLIKE( Nclus,clust_prior,FAIL)
  real*8 :: CALCOSMO_CALCULATE_LNLIKE
  integer, intent(out) :: fail
   real, intent(out) :: Nclus,clust_prior
  real*8 :: data_like,poisson_lnlike,total_expected, prior_lnlike
  real*8, dimension(1) :: this_data_like
  real*8, parameter :: interp_fact = 10d0
  integer, parameter :: iinterp_fact = 10
  integer :: nzspline , nxspline
  real*8, dimension(0:(iinterp_fact * szparams%nz -1)) :: old_z_spline_arr_index,old_z_spline_arr
  real*8, dimension(0:(iinterp_fact * szparams%nx -1)) :: old_x_spline_arr_index,old_x_spline_arr,kernel

  integer :: i_clust, i,zl,zh,isz_index,dnz,zl0,zh0,ifield
  real*8 :: whcutoff,tmpxerr
  real*8 :: delta_z, z_min, this_sz, sz_min,delta_sz,sz_index,ff,this_z,z_index
  real*8, dimension(0:(iinterp_fact * szparams%nz-1)) :: exp_z0, exp_z1, exp_z, exp_z0b, exp_z1b

  real*8, dimension(0:szparams%nz-1) :: zinds
  real*8, dimension(0:szparams%nx-1) :: xinds
  real*8, dimension(0:szparams%nsz-1) :: szinds
  real*8, dimension(1) :: t1,t2
  real*8, dimension(0:3) :: xpriorscr
  real*8, dimension(0:5) :: szpriorscr
  real*8 :: min_x_error,min_z_error, v_term
  real*8 :: x_min,delta_x, this_x, x_index,gg
  integer :: xl, xh,j, ix_index
  logical :: lowerlimit

  CALCOSMO_CALCULATE_LNLIKE=-1d30
  !vikhlinin term
  v_term = 0d0

  nzspline =iinterp_fact * (szparams%nz-1)
  nxspline =iinterp_fact * (szparams%nx-1)
  do i=0,szparams%nz-1
     zinds(i)=i
  end do
  do i=0,szparams%nsz-1
     szinds(i)=i
  end do
  do i=0,szparams%nx-1
     xinds(i)=i
  end do
  fail = 0
  ! check for crazy scaling relations
  if ( sr%ax .lt. 0.001 .or. sr%ax .gt. 100. &
     .or. sr%bx .lt. 0.05 .or. sr%bx .gt. 100. &
     .or. sr%cx .lt. -10. .or. sr%cx .gt. 10. &
     .or. sr%dx .lt. 0.02 .or. sr%dx .gt. 4. &
     .or. sr%asz .lt. 0.001 .or. sr%asz .gt. 100. &
     .or. sr%bsz .lt. 0.05 .or. sr%bsz .gt. 100. &
     .or. sr%csz .lt. -10. .or. sr%csz .gt. 10. &
     .or. sr%fsz .lt. 0.02 .or. sr%fsz .gt. 4. &
     .or. sr%rho .lt. 0. .or. sr%rho .gt. 1. ) then 

     fail = 1
     return
  endif
  
  call calcosmo_generate_grid()

  if (maxval(grid3dexpected) .le. 0) then 
     fail = 1
     write(*,*)'Failed to generate grid'
     return
  endif

  call calcosmo_yx_calc()

  data_like = 0

  delta_sz = (szparams%sz_arr(szparams%nsz-1) - szparams%sz_arr(0)) / (szparams%nsz-1)
  sz_min = szparams%sz_arr(0)

  !want to subdivide x/z array. it's regularly gridded
  delta_z = (szparams%z_arr(szparams%nz-1) - szparams%z_arr(0)) / (szparams%nz-1) / interp_fact
  min_z_error = abs(delta_z /  sqrt(2d0 *Pi))
  z_min = szparams%z_arr(0) 
  do i=0,nzspline-1
     old_z_spline_arr(i) = z_min + delta_z*(i)
  end do
  x_min = szparams%x_arr(0)
  delta_x = (szparams%x_arr(szparams%nx-1) - szparams%x_arr(0)) / (szparams%nx-1) / interp_fact
  min_x_error = abs(delta_x /  sqrt(2d0 *Pi) )
  do i=0,nxspline-1
     old_x_spline_arr(i) = x_min + delta_x*(i)
     old_x_spline_arr_index(i) = (i)/interp_fact
  end do

  total_expected=0d0
  whcutoff = (szparams%szthresh - sz_min) / delta_sz
  do ifield=1,cat%nfield
!$OMP PARALLEL DO DEFAULT(SHARED),SCHEDULE(GUIDED),&
!$OMP REDUCTION(+:data_like,v_term),PRIVATE(i_clust,j,zl,zh,zl0,zh0,xl,xh,tmpxerr,this_sz,sz_index,isz_index,this_x,x_index,ix_index,this_z,z_index,ff,gg,exp_z0,exp_z1,exp_z0b,exp_z1b,exp_z,kernel,t1,t2,this_data_like,dnz,lowerlimit)
     do i_clust=cat%firsti(ifield),cat%firsti(ifield)+cat%nper(ifield)-1 
        lowerlimit = .false.
        if (cat%z_err_vec(i_clust) .le. min_z_error) then	!no z_error
           !assume regular grid for z
           zl = min(max(floor((cat%z_vec(i_clust) - z_min)/delta_z),0),nzspline-1)
           zh = min(zl+1,nzspline-1)
        else
           if (cat%z_err_vec(i_clust) .ge. max_z_error) then
              zl = min(max(floor((cat%z_vec(i_clust) - z_min)/delta_z + 0.5),0),nzspline-1)
              zh = nzspline-1
              lowerlimit = .true.
           else
              zl = min(max(floor((cat%z_vec(i_clust) - z_min - 5*cat%z_err_vec(i_clust))/delta_z),0),nzspline-1)
              zh = min(max(floor((cat%z_vec(i_clust) - z_min + 5*cat%z_err_vec(i_clust))/delta_z),0),nzspline-1)
           end if
        endif
     
        if (cat%x_vec(i_clust) .le. 1d-6) then
           xl = 0
           xh = nxspline-1
        else
           tmpxerr=cat%x_vec(i_clust)*cat%logx_err_vec(i_clust)
           if (tmpxerr .le. min_x_error) then
              xl = min(max(floor((cat%x_vec(i_clust) - x_min)/delta_x),0),nxspline-1)
              xh = min(xl+1,nxspline-1)
           else
              xl = min(max(floor((cat%x_vec(i_clust)/(exp(cat%logx_err_vec(i_clust))**5) - x_min)/delta_x),0),nxspline-1)
              xh = min(max(floor((cat%x_vec(i_clust)*(exp(cat%logx_err_vec(i_clust))**5) - x_min)/delta_x),0),nxspline-1)
           endif
        endif
     
        ! REMOVE ME
        if (zl .eq. zh .or. xl .eq. xh) then 
           fail = 1
           write(*,*) 'Error, data outside of GRID!',i_clust,cat%x_vec(i_clust),x_min,zl,zh,xl,xh
           cycle
        endif
 
        ! 1) Evaluate the grid in dimension 2 (\xi space)
        this_sz = cat%sz_vec(i_clust)
        
        sz_index = (this_sz - sz_min) / delta_sz
        isz_index = floor(sz_index)
        
        !2) Evaluate the grid in dimension 3 (Yx space)
        if (cat%x_vec(i_clust) .le. 1d-6) then
           call dlinterpol(sum(grid3dexpected(:,isz_index,:,ifield),1),szparams%z_arr,szparams%nz,&
                old_z_spline_arr(zl:zh), exp_z0,zh-zl+1)
           call dlinterpol(sum(grid3dexpected(:,isz_index+1,:,ifield),1),szparams%z_arr,szparams%nz,&
                old_z_spline_arr(zl:zh), exp_z1,zh-zl+1)
           ff =  (sz_index-isz_index)
           exp_z(0:zh-zl) = (1d0-ff)*exp_z0(0:zh-zl) + ff*exp_z1(0:zh-zl)
           exp_z(0:zh-zl) = exp_z(0:zh-zl) / (interp_fact *szparams%nx)
        else
           v_term =v_term + log(cat%x_vec(i_clust))
           if (cat%x_vec(i_clust)*cat%logx_err_vec(i_clust) .le. min_x_error) then
              !no Yx error (simulated data)
              this_x = cat%x_vec(i_clust)
              x_index = (this_x - x_min) / (delta_x*interp_fact)
              ix_index = floor(x_index)
              
              call dlinterpol(grid3dexpected(ix_index,isz_index,:,ifield),szparams%z_arr,&
                   szparams%nz,&
                   old_z_spline_arr(zl:zh), exp_z0,zh-zl+1)
              call dlinterpol(grid3dexpected(ix_index,isz_index+1,:,ifield),szparams%z_arr,&
                   szparams%nz,&
                   old_z_spline_arr(zl:zh), exp_z1,zh-zl+1)
              call dlinterpol(grid3dexpected(ix_index+1,isz_index,:,ifield),szparams%z_arr,&
                   szparams%nz,&
                   old_z_spline_arr(zl:zh), exp_z0b,zh-zl+1)
              call dlinterpol(grid3dexpected(ix_index+1,isz_index+1,:,ifield),szparams%z_arr,&
                   szparams%nz,&
                   old_z_spline_arr(zl:zh), exp_z1b,zh-zl+1)
              ff =  (sz_index-isz_index)
              gg =  (x_index-ix_index)
              exp_z(0:zh-zl) = (1-gg)*((1d0-ff)*exp_z0(0:zh-zl) + ff*exp_z1(0:zh-zl)) +&
                   gg*((1d0-ff)*exp_z0b(0:zh-zl) + ff*exp_z1b(0:zh-zl))
              exp_z(0:zh-zl) = exp_z(0:zh-zl) / interp_fact
           else
           ! Yx measurement, use a kernel centered at the best fit Yx, with Gaussian width of the measurement error
              exp_z(0:zh-zl) = 0
              kernel(0:xh-xl) = exp(-0.5 * (&
                   (log(old_x_spline_arr(xl:xh)) - log(cat%x_vec(i_clust)))/&
                   cat%logx_err_vec(i_clust) )**2) / &
                   (sqrt(2d0*Pi * cat%logx_err_vec(i_clust)**2 ) * &
                   old_x_spline_arr(xl:xh))
              gg=sum(kernel(0:xh-xl))
              kernel(0:xh-xl) = kernel(0:xh-xl) / gg
              
              do j=xl,xh
                 this_x = old_x_spline_arr(j)
                 x_index = (this_x - x_min) / (delta_x*interp_fact)
                 ix_index = floor(x_index)
                 call dlinterpol(grid3dexpected(ix_index,isz_index,:,ifield),szparams%z_arr,&
                      szparams%nz,&
                      old_z_spline_arr(zl:zh), exp_z0,zh-zl+1)
                 call dlinterpol(grid3dexpected(ix_index,isz_index+1,:,ifield),szparams%z_arr,&
                      szparams%nz,&
                      old_z_spline_arr(zl:zh), exp_z1,zh-zl+1)
                 call dlinterpol(grid3dexpected(ix_index+1,isz_index,:,ifield),szparams%z_arr,&
                      szparams%nz,&
                      old_z_spline_arr(zl:zh), exp_z0b,zh-zl+1)
                 call dlinterpol(grid3dexpected(ix_index+1,isz_index+1,:,ifield),szparams%z_arr,&
                      szparams%nz,&
                      old_z_spline_arr(zl:zh), exp_z1b,zh-zl+1)
                 ff =  (sz_index-isz_index)
                 gg =  (x_index-ix_index)
                 exp_z(0:zh-zl) = exp_z(0:zh-zl) + kernel(j-xl) *((1-gg)*(&
                      (1d0-ff)*exp_z0(0:zh-zl) + ff*exp_z1(0:zh-zl)) +&
                      gg*((1d0-ff)*exp_z0b(0:zh-zl) + ff*exp_z1b(0:zh-zl)))
              enddo
              exp_z(0:zh-zl) = exp_z(0:zh-zl)/interp_fact
           endif
        endif
        
  ! 2) Evaluate the grid in dimension 1 (z space)
        this_z = cat%z_vec(i_clust)
        z_index = (this_z - z_min) / delta_z - zl
        if (cat%z_err_vec(i_clust) .le. min_z_error) then
        ! no z error (spec-z, simulated data)
           t1(1) = z_index
           call dlinterpol(exp_z(0:zh-zl),zinds(0:zh-zl),zh-zl+1,t1,t2,1)
           this_data_like = t2(1) /interp_fact 
        ! dividing by extra interp_fact since will divide by 1/interp later
        else  
           if (lowerlimit) then 
              !summing everything up
              this_data_like(1) = sum(exp_z(0:zh-zl))/interp_fact  + szparams%falserates(isz_index,ifield)/interp_fact 
        ! dividing by extra interp_fact since will divide by 1/interp later
           else
              dnz = floor(5*(cat%z_err_vec(i_clust)/delta_z))+1
              zl0 = max(floor(z_index-dnz),0)+zl
              zh0 = min(floor(z_index+dnz),zh-zl)+zl
              exp_z0(0:zh0-zl0) = (old_z_spline_arr(zl0:zh0) - this_z)/cat%z_err_vec(i_clust)
              exp_z0(0:zh0-zl0) = exp(-0.5 * (exp_z0(0:zh0-zl0))**2)
              exp_z0(0:zh0-zl0) = exp_z0(0:zh0-zl0) /sum(exp_z0(0:zh0-zl0))
              this_data_like(1) = sum(exp_z0(0:zh0-zl0) *exp_z(zl0-zl:zh0-zl)) /interp_fact
           ! this is for bin size
           endif
        endif
        
        ! we can rescale by binsize, which only adds a constant to the lnlike
        !   however, the lnlike now becomes independent of binsize
        this_data_like = this_data_like/(delta_z*delta_sz*delta_x)

        data_like = data_like + log(this_data_like(1))
     end do
  !$OMP END PARALLEL DO
  
  ! now use a fancy 3d integrator (canned routine)
     total_expected = total_expected + &
          cube_int_3d(grid3dexpected(:,:,:,ifield), xinds,szparams%nx,&
          szinds,szparams%nsz,zinds,szparams%nz,&
          (/ 0d0, 1d0*szparams%nx-1d0, 1d0*whcutoff, 1d0*szparams%nsz-1, &
          0d0, 1d0*szparams%nz-1 /), 20)
  enddo
  total_expected = total_expected+sum(szparams%falserates(floor(whcutoff+1d-6):szparams%nsz-1,1:cat%nfield))

  Nclus = total_expected
    
  poisson_lnlike = data_like - total_expected

  ! add X and SZ priors when applicable
  prior_lnlike=0.d0
  szpriorscr(:)=0
  xpriorscr(:)=0
  where (szparams%x_prior > 0.0)
     xpriorscr = -0.5 * ((szparams%x_central_vals - &
          (/ sr%ax,sr%bx,sr%cx,sr%dx /))/szparams%x_prior)**2
  endwhere
  where (szparams%sz_prior > 0.0)
     szpriorscr = -0.5 * ( (szparams%sz_central_vals - &
          (/ sr%asz,sr%bsz,sr%csz,sr%dsz,sr%esz,sr%fsz /) ) / &
          szparams%sz_prior)**2
  endwhere

#ifdef _NOPRIOR_
 prior_lnlike = prior_lnlike 
#else
  prior_lnlike = prior_lnlike + sum(xpriorscr) + sum(szpriorscr)
#endif

  clust_prior=-1*prior_lnlike

  prior_lnlike = prior_lnlike + v_term
  CALCOSMO_CALCULATE_LNLIKE = -1*(prior_lnlike + poisson_lnlike)
 
END function CALCOSMO_CALCULATE_LNLIKE



SUBROUTINE CALCOSMO_GENERATE_GRID()
  use MKL_VSL
  use MKL_VSL_TYPE
  integer, save :: firsttime = 1
  integer :: ifield
  integer, parameter :: MAXSIZE = 1024
  real*8, dimension(0:MAXSIZE-1), save :: lnszdb,  alogm,szdb_arr,lnx
  real*8, dimension(0:szparams%nx-1) ::  x_masses,sigx_masses,&
       dev_sigma_sq
  real*8, dimension(0:szparams%nm-1) :: cen_lnmsz,cen_lnmx, censz,ocensz,cenx,ocenx,&
       g2,tmpnm,tmp2nm
  real*8, dimension(0:szparams%nsz-1) :: junknsz, junk2nsz,junk3nsz,&
       sz_masses,sigsz_masses
  real*8, save :: sz_binsize, x_binsize, z_binsize
  real*8, parameter :: sigma = 6d0  ! IMPORTANT APPROXIMATION PARAMETER
  real*8 :: expdsz , sigsz,  ll_sz, ul_sz, denom_exp_arg, renorm_fact
  real*8 :: ll_x, ul_x, cen_ul, cen_ll
  integer, parameter :: nz_int = 1000
  real*8, dimension(0:nz_int) :: z_int_arr, iez_int_arr
  real*8, dimension(0:mf%nz-1) :: da_arr, ez_arr
  real*8 :: da_pivot, ez_pivot, sk, sz_sigma_obs, D_H
  real*8 :: delta_cen_lnmx, delta_cen_lnmsz
  integer :: i,j,k
  integer :: k0,k1,j0,j1,nuse
  !conv task vars
  TYPE(VSL_CONV_TASK), save ::  task, task2
  integer start(1)
  real*8 :: x(9),h(9),y(18)
  integer ::  mode,xshape,yshape,zshape,xstride
  integer, save :: nkern,halfnkern
  integer :: status
  integer, parameter :: MAXNKERN = 1024
  real*8, save, dimension(MAXNKERN) :: ker, zout,zcor
  real*8, dimension(1) :: t1,t2

  !define 'de-boosted' sz vector
  if (firsttime .eq. 1) then
     if ((MAXSIZE .lt. szparams%nsz) .or. (MAXSIZE .lt. szparams%nm)) then
        write(*,*) 'maxsize too small',MAXSIZE, szparams%nsz,szparams%nm
        call DoStop
     endif
     firsttime = 0
     do i=0,szparams%nsz-1
        if (szparams%sz_arr(i) .gt. sqrt(7d0)) then
           szdb_arr(i) = sqrt(szparams%sz_arr(i)**2 - 3d0)
        else
           szdb_arr(i) = szparams%sz_arr(i)
        end if
        lnszdb(i) = log(szdb_arr(i))

     enddo

     lnx(0:szparams%nx-1) = log(szparams%x_arr(0:szparams%nx-1))
     sz_binsize = abs(szparams%sz_arr(1) - szparams%sz_arr(0))
     z_binsize = szparams%z_arr(1) - szparams%z_arr(0)
     x_binsize = szparams%x_arr(1) - szparams%x_arr(0)
     alogm(0:szparams%nm-1) = log(szparams%m_arr(:)/szparams%mpivot_sz)


     sz_sigma_obs=1d0/sz_binsize
     halfnkern = floor(5*sz_sigma_obs)+1

     nkern = 2*halfnkern+1

     if (nkern + szparams%nz -1 .gt. MAXNKERN) then
        write(*,*)'MAXNKERN too small',nkern + szparams%nz -1, MAXNKERN
        call DoStop
     endif
     do i=-1*halfnkern, halfnkern
        ker(1+i+halfnkern) = exp(-0.5/sz_sigma_obs**2 * (i**2) )
     end do
     sk=sum(ker(1:nkern))
     ker(1:nkern) = ker(1:nkern)/sk
     zcor(halfnkern) =  ker(1)
     do i=2,halfnkern
        zcor(halfnkern+1 - i) = zcor(halfnkern+2 - i) + ker(i)
     end do

     mode = VSL_CONV_MODE_DIRECT;  
     status = vsldconvnewtaskx1d(task, mode, nkern, szparams%nsz, szparams%nsz, ker, 1)

     start(1) = halfnkern
     status = vslconvsetstart(task,start) 
     
  endif

  ll_sz = -1*sigma * sr%fsz
  ul_sz = sigma * sr%fsz
  ll_x = -1*sigma * sr%dx
  ul_x = sigma * sr%dx

  D_H=2997.92458/(csm%H0/100.0)		!Hubble distance, note factor of h!
  grid3dexpected(:,:,:,:) = 0

  sigsz = 2d0*sr%fsz**2 
  denom_exp_arg = 2d0*sr%fsz**2 * sr%dx**2 * (1d0 - sr%rho**2)
  renorm_fact = x_binsize * sz_binsize / &
       (2d0*Pi*sr%fsz*sr%dx*sqrt(1d0-sr%rho**2))

  !create angular diameter distance array (note units here are MPC, not h**(-1)MPC)
  ez_arr(:) = sqrt( csm%omm*(1d0+szparams%z_arr(:))**3 + (csm%oml) * (1d0+szparams%z_arr(:))**(3d0*(1d0+(csm%w))))
  
  do i=0,szparams%nz-1 
     do j=0,nz_int
        z_int_arr(j) = j*(szparams%z_arr(i)/nz_int)
     end do
     
     iez_int_arr(:) = 1d0 / sqrt( csm%omm*(1d0 + z_int_arr(:))**3d0 + (csm%oml) * (1d0 + z_int_arr(:))**(3d0 * (1d0 + (csm%w)) ) )
     da_arr(i)=D_H*Dint_tabulated_reg(z_int_arr,iez_int_arr,nz_int+1)/(1.0+szparams%z_arr(i))
  end do

!calculate pivot points for E(z) and D_A(z)
  t1(1) = szparams%z_pivot
  call dsinterpol(ez_arr,szparams%z_arr,mf%nz, t1,t2, 1)
  ez_pivot = t2(1)
  call dsinterpol(da_arr,szparams%z_arr,mf%nz, t1,t2, 1)
  da_pivot = t2(1)

  junknsz(:) =  renorm_fact/szparams%sz_arr(:)
  
 !populate sz-z grid
  !both arrays below are uniformly spaced.
  delta_cen_lnmsz =  sr%bsz *(alogm(szparams%nm-1)-alogm(0))/(szparams%nm-1)
  delta_cen_lnmx =  (1d0/sr%bx) *(alogm(szparams%nm-1)-alogm(0))/(szparams%nm-1)
  

  !$OMP PARALLEL DO DEFAULT(SHARED),PRIVATE(cen_lnmsz,cen_lnmx,i,j,j0,j1,k,k0,k1,nuse,censz,ocensz,cenx,ocenx,g2,tmpnm,tmp2nm,junk2nsz,zout,status,ifield),SCHEDULE(GUIDED)
  do i=0,mf%nz-1 
     
     if (maxval(mf%dn(:,i)) .eq. 0) cycle ! skip empty redshift bins
     
     cen_lnmsz(:) = log(sr%asz) + sr%bsz * alogm(0:szparams%nm-1) + &
          sr%csz * log((1d0+szparams%z_arr(i))/(1d0+szparams%z_pivot))+ &
          sr%dsz * log(ez_arr(i)/ez_pivot)+ sr%esz * log(da_arr(i)/da_pivot)
     cen_lnmx(:) = log(3d0) + (alogm(0:szparams%nm-1) - &
          log(sr%ax / (szparams%mpivot_sz/1d14 ) * (csm%h0/100d0)**1.5) - &
sr%cx * log(ez_arr(i))) / sr%bx

     do j=0,szparams%nsz-1 
        if (szparams%sz_arr(j) .lt. sqrt(7d0) .and. &
             szparams%sz_arr(j) .ge. 2d0) then
            cycle
        endif
        
        do ifield=1,cat%nfield
           ocensz = lnszdb(j) - cen_lnmsz  - log(cat%szscale(ifield))
           !define j0,j1 - indices over which to calculate
           !cen_lnmsz is increasing, so ocensz is monotonically decreasing
           !0, or first index under upper limit
           j0 = max(0,min(szparams%nm-1,floor((ocensz(0)-ul_sz) / delta_cen_lnmsz)))
           ! and nm-1 or first index under lower limit
           j1 = max(0,min(szparams%nm-1,floor((ocensz(0)-ll_sz) / delta_cen_lnmsz)))
           
           nuse = j1-j0+1
           if (nuse .le. 0) cycle
           
           call vdsqr(szparams%nm,ocensz, censz)
           censz = (sr%dx**2) * censz
           
           do k=0,szparams%nx-1
              !don't need both 0,nm-1 bounds since j0.j1 bounded
              !finding intersection of two useable bounds
              k0 = min(szparams%nm-1,max(j0,floor((lnx(k)-cen_lnmx(0)-ul_x)/delta_cen_lnmx)))
              k1 = max(0,min(j1,floor((lnx(k)-cen_lnmx(0)-ll_x)/delta_cen_lnmx)))
              nuse = k1-k0+1
              if (nuse .le. 0) cycle
              ocenx(k0:k1) = lnx(k) - cen_lnmx(k0:k1)
              
              call vdsqr(nuse,ocenx(k0:k1), cenx(k0:k1))
              cenx(k0:k1) = cenx(k0:k1) * (sr%fsz**2)
 
              call vdmul(nuse,ocenx(k0:k1),ocensz(k0:k1),g2(k0:k1))
              g2(k0:k1) = g2(k0:k1) * (-2d0 * sr%rho * sr%dx * sr%fsz)
               
              !sum
              call vdadd(nuse,cenx(k0:k1),g2(k0:k1),tmpnm(k0:k1))
              
              call vdadd(nuse,censz(k0:k1),tmpnm(k0:k1),tmp2nm(k0:k1))
                 
              tmp2nm(k0:k1) = (-1d0/denom_exp_arg) * tmp2nm(k0:k1)
              call vdexp(nuse,tmp2nm(k0:k1),tmpnm(k0:k1))
              call vdmul(nuse,mf%dn(k0:k1,i),tmpnm(k0:k1),tmp2nm(k0:k1))
              grid3dexpected(k,j,i,ifield) = sum(tmp2nm(k0:k1)) * cat%fracarea(ifield)
           enddo ! k
        enddo ! ifield
     enddo ! j
     do ifield=1,cat%nfield
        do k=0,szparams%nx-1
        
           call vdmul(szparams%nsz,grid3dexpected(k,:,i,ifield),junknsz,junk2nsz)
        
           status = vsldconvexecx1d(task, junk2nsz, 1, zout, 1)
           zout(1:halfnkern) = zout(1:halfnkern) +& 
                junk2nsz(0) * zcor(1:halfnkern)
           !upper edge
           zout(szparams%nsz:szparams%nsz-halfnkern+1:-1) = &
                zout(szparams%nsz:szparams%nsz-halfnkern+1:-1) +&
                junk2nsz(szparams%nsz-1)* zcor(1:halfnkern)
           
           !get 1/x factor
           zout(1:szparams%nsz) = zout(1:szparams%nsz) / szparams%x_arr(k)
           call vdpacki(szparams%nsz,zout,1,grid3dexpected(k,:,i,ifield))
        enddo
     enddo
  enddo
!$OMP END PARALLEL DO

END SUBROUTINE CALCOSMO_GENERATE_GRID


subroutine calcosmo_yx_calc
  implicit none
  integer, parameter :: nz_int = 1000
  real*8, dimension(0:nz_int) :: z_int_arr, iez_int_arr
  real*8, dimension(0:mf%nz-1) :: da_arr, ez_arr
  real*8 :: da_pivot, ez_pivot
  real*8 :: D_H
  real*8, dimension(0:MAXNCATALOG-1) :: da_vec,ez_vec,yx_vec
  integer :: i,j,maxind2,maxind
  real*8 :: da_ref,r500,da,ez,z,yxold,yxtol
  real*8 :: r500ref,angrad
  real*8 :: M500, rhoc, Yx, Mg, kT
  real*8, parameter :: G = 4.3d-3
  real*8, dimension(1) :: t1, t2

   D_H = 2997.92458/(csm%H0/100.0)

  ez_arr(:) = sqrt( csm%omm*(1d0+szparams%z_arr(:))**3 + (csm%oml) * (1d0+szparams%z_arr(:))**(3d0*(1d0+(csm%w))))
  do i=0,szparams%nz-1 
     do j=0,nz_int
        z_int_arr(j) = j*(szparams%z_arr(i)/nz_int)
     end do
     iez_int_arr(:) = 1d0 / sqrt( csm%omm*(1d0 + z_int_arr(:))**3d0 + (csm%oml) * (1d0 + z_int_arr(:))**(3d0 * (1d0 + (csm%w)) ) )
     da_arr(i)=D_H*Dint_tabulated_reg(z_int_arr,iez_int_arr,nz_int+1)/(1.0+szparams%z_arr(i))
  end do

!calculate pivot points for E(z) and D_A(z)
  t1(1) = szparams%z_pivot
  call dsinterpol(ez_arr,szparams%z_arr,mf%nz, t1,t2, 1)
  ez_pivot = t2(1)
  call dsinterpol(da_arr,szparams%z_arr,mf%nz, t1,t2, 1)
  da_pivot = t2(1)

  !init first to set extra elements to zero just in case
  yx_vec(:)=0d0
  ez_vec(:)=0d0
  da_vec(:)=0d0
  call dsinterpol(da_arr,szparams%z_arr,mf%nz,cat%z_vec,da_vec,cat%n)
  call dsinterpol(ez_arr,szparams%z_arr,mf%nz,cat%z_vec,ez_vec,cat%n)

  !cluster loop
  do i=0,cat%n-1 
     da_ref = cat%da_vec(i)
     r500=1000.0
     !check for no xray
     if (da_ref .le. 0.0 .or. &
          maxval(cat%mgr_vec(:,i)) .le. 0. .or. &
          maxval(cat%txr_vec(:,i)) .le. 0. ) then
        yx_vec(i)=-1e14
        cycle !goto next cluster
     endif
     
!code to cut to nelements here
     maxind=cat%nmg-1
     do j=cat%nmg-1,0,-1
        if (cat%mgr_vec(j,i) .gt. 0 .and. cat%r_mg(j,i) .gt. 0 ) then 
           maxind = j
           exit 
        endif
     enddo
     maxind2=cat%ntx-1
     do j=cat%nmg-1,0,-1
        if (cat%txr_vec(j,i) .gt. 0 .and. cat%r_tx(j,i) .gt. 0 ) then 
           maxind2 = j
           exit 
        endif
     enddo
          
     z = cat%z_vec(i)
     da = da_vec(i)
     ez = ez_vec(i)
     yxold=0
     yxtol = 1e10
     do j=0,99 
        r500ref=r500/da*da_ref
        angrad=r500/da/1000./Pi*180.*3600.
        !define Mg/kT here WITH INTERPOL
        t1(1)=r500ref
        call dsinterpol(cat%mgr_vec(0:maxind,i),cat%r_mg(0:maxind,i),maxind+1,t1,t2,1)
        Mg = t2(1)
        t1(1)=angrad
        call dsinterpol(cat%txr_vec(0:maxind2,i),cat%r_tx(0:maxind2,i),maxind2+1,t1,t2,1)
        kT = t2(1)
        
        Mg=Mg/da_ref**2.5*da**2.5

        Yx=Mg*kT
        M500=SR%ax*1d14*((csm%H0/100.0)**0.5)*(Yx/(3d14))**(SR%bx)*ez**(SR%cx) !true iff SR.bx==3/5         
        rhoc=3.*(csm%H0*ez)**2 / (8.*Pi*G) * 1d6
        r500=1000.*(M500/rhoc/500.*3./4./Pi)**(1./3.)
        if (abs(Yx-yxold) .lt. yxtol .and. j .gt. 9)  exit
        yxold=Yx
     enddo
     yx_vec(i)=Yx

  enddo
  
  CAT%x_vec(0:cat%n-1)=yx_vec(0:cat%n-1)/1d14
  return
end subroutine calcosmo_yx_calc

end module szcluster

  
